@@ -62,7 +62,7 @@ class btchip:
 	BTCHIP_INS_EXT_CACHE_GET_FEATURES = 0x26
 
 	OPERATION_MODE_WALLET = 0x01
-	OPERATION_MODE_RELAXED_WALLET = 0x02
+	OPERATION_MODE_RELAXED_WALLET = 0x02 
 	OPERATION_MODE_SERVER = 0x04
 	OPERATION_MODE_DEVELOPER = 0x08
 
@@ -79,6 +79,15 @@ class btchip:
 		self.dongle = dongle
 		self.needKeyCache = False
 		try:
+			firmware = self.getFirmwareVersion()['version'].split(".")
+			self.multiOutputSupported = int(firmware[0]) >= 1 and int(firmware[1]) >= 1 and int(firmware[2]) >= 4
+			if self.multiOutputSupported:
+				self.scriptBlockLength = 50
+			else:
+				self.scriptBlockLength = 255
+		except:
+			pass				
+		try:			
 			result = self.getJCExtendedFeatures()
 			self.needKeyCache = (result['proprietaryApi'] == False)
 		except:
@@ -159,6 +168,7 @@ class btchip:
 		apdu.extend(params)
 		self.dongle.exchange(bytearray(apdu))
 		# Each output
+		indexOutput = 0
 		for troutput in transaction.outputs:
 			apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_GET_TRUSTED_INPUT, 0x80, 0x00 ]
 			params = bytearray(troutput.amount)
@@ -187,7 +197,20 @@ class btchip:
 
 	def startUntrustedTransaction(self, newTransaction, inputIndex, outputList, redeemScript):
 		# Start building a fake transaction with the passed inputs
-		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_HASH_INPUT_START, 0x00, (0x00 if newTransaction else 0x80) ]
+		segwit = False
+		if newTransaction:
+			for passedOutput in outputList:
+				if ('witness' in passedOutput) and passedOutput['witness']:
+					segwit = True
+					break
+		if newTransaction:
+			if segwit:
+				p2 = 0x02
+			else:
+				p2 = 0x00
+		else:
+				p2 = 0x80
+		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_HASH_INPUT_START, 0x00, p2 ]
 		params = bytearray([0x01, 0x00, 0x00, 0x00]) # default version
 		writeVarint(len(outputList), params)
 		apdu.append(len(params))
@@ -199,11 +222,13 @@ class btchip:
 			apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_HASH_INPUT_START, 0x80, 0x00 ]
 			params = []
 			script = redeemScript
-			if passedOutput['trustedInput']:
+			if ('witness' in passedOutput) and passedOutput['witness']:
+				params.append(0x02)
+			elif ('trustedInput' in passedOutput) and passedOutput['trustedInput']:
 				params.append(0x01)
 			else:
 				params.append(0x00)
-			if passedOutput['trustedInput']:
+			if ('trustedInput' in passedOutput) and passedOutput['trustedInput']:
 				params.append(len(passedOutput['value']))
 			params.extend(passedOutput['value'])
 			if currentIndex <> inputIndex:
@@ -236,6 +261,7 @@ class btchip:
 		if self.needKeyCache:
 			self.resolvePublicKeysInPath(changePath)		
 		result = {}
+		outputs = None
 		if rawTx is not None:
 			try:
 				fullTx = bitcoinTransaction(bytearray(rawTx))
@@ -247,12 +273,20 @@ class btchip:
 					apdu.append(len(params))
 					apdu.extend(params)
 					response = self.dongle.exchange(bytearray(apdu))
-				apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_HASH_INPUT_FINALIZE_FULL, 0x80, 0x00 ]
-				params = []
-				params.extend(outputs)
-				apdu.append(len(params))
-				apdu.extend(params)
-				response = self.dongle.exchange(bytearray(apdu))
+				offset = 0
+				while (offset < len(outputs)):
+					blockLength = self.scriptBlockLength
+					if ((offset + blockLength) < len(outputs)):
+						dataLength = blockLength
+						p1 = 0x00
+					else:
+						dataLength = len(outputs) - offset
+						p1 = 0x80
+					apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_HASH_INPUT_FINALIZE_FULL, \
+						p1, 0x00, dataLength ]
+					apdu.extend(outputs[offset : offset + dataLength])
+					response = self.dongle.exchange(bytearray(apdu))
+					offset += dataLength
 				alternateEncoding = True
 			except:
 				pass
@@ -278,7 +312,14 @@ class btchip:
 			result['keycardData'] = response[offset : offset + keycardDataLength]
 			offset = offset + keycardDataLength
 			result['secureScreenData'] = response[offset:]
-		result['outputData'] = response[1 : 1 + response[0]]
+		if result['confirmationType'] == 0x04:
+			offset = 1 + response[0] + 1
+			keycardDataLength = response[offset]
+			result['keycardData'] = response[offset + 1 : offset + 1 + keycardDataLength]			
+		if outputs == None:
+			result['outputData'] = response[1 : 1 + response[0]]
+		else:
+			result['outputData'] = outputs
 		return result
 
 	def finalizeInputFull(self, outputData):
@@ -286,7 +327,7 @@ class btchip:
 		offset = 0
 		encryptedOutputData = ""
 		while (offset < len(outputData)):
-			blockLength = 255
+			blockLength = self.scriptBlockLength
 			if ((offset + blockLength) < len(outputData)):
 				dataLength = blockLength
 				p1 = 0x00
@@ -299,20 +340,27 @@ class btchip:
 			response = self.dongle.exchange(bytearray(apdu))
 			encryptedOutputData = encryptedOutputData + response[1 : 1 + response[0]]
 			offset += dataLength
-                result['confirmationNeeded'] = response[1 + response[0]] <> 0x00
-                result['confirmationType'] = response[1 + response[0]]
-                if result['confirmationType'] == 0x02:
+		if len(response) > 1: 
+			result['confirmationNeeded'] = response[1 + response[0]] <> 0x00
+			result['confirmationType'] = response[1 + response[0]]
+		else:
+			# Support for old style API before 1.0.2
+			result['confirmationNeeded'] = response[0] <> 0x00
+			result['confirmationType'] = response[0]
+		if result['confirmationType'] == 0x02:
+			result['keycardData'] = response[1 + response[0] + 1:] # legacy
+		if result['confirmationType'] == 0x03:
 			offset = 1 + response[0] + 1
 			keycardDataLength = response[offset]
-                        result['keycardData'] = response[offset : offset + keycardDataLength]
-                if result['confirmationType'] == 0x03:
-                        offset = 1 + response[0] + 1
-                        keycardDataLength = response[offset]
-                        offset = offset + 1
-                        result['keycardData'] = response[offset : offset + keycardDataLength]
-                        offset = offset + keycardDataLength
-                        result['secureScreenData'] = response[offset:]
-                result['encryptedOutputData'] = encryptedOutputData 
+			offset = offset + 1
+			result['keycardData'] = response[offset : offset + keycardDataLength]
+			offset = offset + keycardDataLength
+			result['secureScreenData'] = response[offset:]
+			result['encryptedOutputData'] = encryptedOutputData 
+		if result['confirmationType'] == 0x04:
+			offset = 1 + response[0] + 1
+			keycardDataLength = response[offset]
+			result['keycardData'] = response[offset + 1 : offset + 1 + keycardDataLength]						
 		return result
 
 	def untrustedHashSign(self, path, pin="", lockTime=0, sighashType=0x01):
@@ -332,7 +380,7 @@ class btchip:
 		result[0] = 0x30
 		return result
 
-	def signMessagePrepare(self, path, message):
+	def signMessagePrepareV1(self, path, message):
 		donglePath = parse_bip32_path(path)
 		if self.needKeyCache:
 			self.resolvePublicKeysInPath(path)		
@@ -347,10 +395,57 @@ class btchip:
 		response = self.dongle.exchange(bytearray(apdu))
 		result['confirmationNeeded'] = response[0] <> 0x00
 		result['confirmationType'] = response[0]
-                if result['confirmationType'] == 0x02:
-                        result['keycardData'] = response[1:]
-                if result['confirmationType'] == 0x03:
-                        result['secureScreenData'] = response[1:]
+		if result['confirmationType'] == 0x02:
+			result['keycardData'] = response[1:]
+		if result['confirmationType'] == 0x03:
+			result['secureScreenData'] = response[1:]
+		return result
+
+	def signMessagePrepareV2(self, path, message):
+		donglePath = parse_bip32_path(path)
+		if self.needKeyCache:
+			self.resolvePublicKeysInPath(path)				
+		result = {}
+		offset = 0
+		encryptedOutputData = ""
+		while (offset < len(message)):
+			params = [];
+			if offset == 0:
+				params.extend(donglePath)
+				params.append((len(message) >> 8) & 0xff)
+				params.append(len(message) & 0xff)
+				p2 = 0x01
+			else:
+				p2 = 0x80
+			blockLength = 255 - len(params)
+			if ((offset + blockLength) < len(message)):
+				dataLength = blockLength
+			else:
+				dataLength = len(message) - offset
+			params.extend(bytearray(message[offset : offset + dataLength]))
+			apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_SIGN_MESSAGE, 0x00, p2 ]
+			apdu.append(len(params))
+			apdu.extend(params)
+			response = self.dongle.exchange(bytearray(apdu))
+			encryptedOutputData = encryptedOutputData + response[1 : 1 + response[0]]
+			offset += blockLength
+		result['confirmationNeeded'] = response[1 + response[0]] <> 0x00
+		result['confirmationType'] = response[1 + response[0]]
+		if result['confirmationType'] == 0x03:
+			offset = 1 + response[0] + 1
+			result['secureScreenData'] = response[offset:]			
+			result['encryptedOutputData'] = encryptedOutputData 
+
+		return result
+
+	def signMessagePrepare(self, path, message):
+		try:
+			result = self.signMessagePrepareV2(path, message)
+		except BTChipException as e:
+			if (e.sw == 0x6b00): # Old firmware version, try older method
+				result = self.signMessagePrepareV1(path, message)
+			else:
+				raise
 		return result
 
 	def signMessageSign(self, pin=""):
@@ -397,11 +492,11 @@ class btchip:
 		self.setKeymapEncoding(keymapEncoding)
 		try:
 			self.setTypingBehaviour(0xff, 0xff, 0xff, 0x10)	
-                except BTChipException as e:
-                        if (e.sw == 0x6700): # Old firmware version, command not supported
-                                pass
-                        else:
-                                raise
+		except BTChipException as e:
+			if (e.sw == 0x6700): # Old firmware version, command not supported
+				pass
+			else:
+				raise
 		return result
 
 	def setKeymapEncoding(self, keymapEncoding):
@@ -435,6 +530,14 @@ class btchip:
 		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_SET_OPERATION_MODE, 0x00, 0x00, 0x01, operationMode ]
 		self.dongle.exchange(bytearray(apdu))
 
+	def enableAlternate2fa(self, persistent):
+		if persistent:
+			p1 = 0x02
+		else:
+			p1 = 0x01
+		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_SET_OPERATION_MODE, p1, 0x00, 0x01, btchip.OPERATION_MODE_WALLET ]
+		self.dongle.exchange(bytearray(apdu))
+
 	def getFirmwareVersion(self):
 		result = {}
 		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_GET_FIRMWARE_VERSION, 0x00, 0x00, 0x00 ]
